/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

#ifndef ENET_STREAM_COMMON_H_
#define ENET_STREAM_COMMON_H_

#include "enet_module.h"
#include "enet_socket.h"
#include "enet_stream.h"
#include "enet_stream_cong.h"
#include "enet_compat.h"

//#if ENET_DEBUG
static const char *esp_statenames[] = {
	"UNUSED", "ESTABLISHED", "SYN SENT", "SYN RECV",
	"FIN WAIT 1", "FIN WAIT 2", "TIME WAIT", "CLOSE",
	"CLOSE WAIT", "LAST ACK", "LISTEN", "CLOSING"
};
//#endif

/* internal function prototypes */

void esp_rtx_timer_finwait(struct sock *sk);
void esp_rtx_timer(unsigned long data);
void esp_dack_timer(unsigned long data);
void esp_sk_destroy(struct sock *sk);   /* no more inline - gcc 4 can't do so */
void sock_esp_destruct(struct sock *sk);

/*
 * Remove the given socket from the port hashtable
 */
static inline void esp_put_port(struct sock *sk)
{
	struct esp_opt *esp_sk = ESP_SK(sk);

	PR_DEBUG("esp socket %p is releasing port %d", sk, esp_sk->port);

	write_lock_bh(&esp_bhash_lock);
	sk_del_bind_node_init(sk);
	esp_sk->port    = 0;
	esp_sk->lport   = 0;
	esp_sk->ifindex = 0;
	esp_sk->bound   = 0;
	write_unlock_bh(&esp_bhash_lock);
}

/*
 * Remove the given socket from the listen/established hashtable
 */
static inline void esp_sk_unhash(struct sock *sk)
{
	if (!sk_hashed(sk)) {
		printk("esp_sk_unhash called for unhashed socket %p", sk);
		return;
	}
	switch (sk->sk_state) {
	case ESP_LISTEN:
		printk("unhashing esp socket %p from listen hashtable\n", sk);
		write_lock_bh(&esp_lhash_lock);
		__sk_del_node_init(sk);
		write_unlock_bh(&esp_lhash_lock);
		break;

	default:
		printk("unhashing esp socket %p from established hashtable\n", sk);
		write_lock_bh(&esp_ehash_lock);
		__sk_del_node_init(sk);
		write_unlock_bh(&esp_ehash_lock);
		break;
	}
}

/*
 * Central esp socket state changing function
 */
static inline void esp_set_state(struct sock *sk, int state)
{

	printk("esp socket %p changing state %s -> %s\n", sk,
	       esp_statenames[sk->sk_state], esp_statenames[state]);

	/* state-dependent operations */
	switch (state) {
	case ESP_FIN_WAIT2:
		esp_rtx_timer_finwait(sk);
		esp_rrq_clear(sk);
		break;
	case ESP_CLOSE:
		/* remove the socket from any request queue */
		esp_openreq_dequeue(sk);
		/* stop timers */
		esp_rrq_clear(sk);
		sk_stop_timer(sk, &ESP_SK(sk)->rtx_timer);

		/* clear ACK queue */
		while (esp_purge_ack(sk));

		/* decrement recv. count */
		if (ESP_SK(sk)->recv_announced) {
			if (!ESP_SK(sk)->rcv_mem_pressure) {
				atomic_dec(&esp_recv_count);
			}
		}

		/* remove from listen/established hashtables */
		if (sk_hashed(sk))
			esp_sk_unhash(sk);

		/* release assigned local port */
		if (sk_bound(sk))
			esp_put_port(sk);
		break;
	}

	sk->sk_state = state;
}

/*
 * convenience function for skb allocation
 */
static inline struct sk_buff *esp_skb_alloc_sk(struct sock *sk,
			                struct net_device *dev,
			                size_t data_len, unsigned char flags,
			                gfp_t allocation, int *err) {
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct sk_buff *skb;
	struct esp_header *esp_hdr;

	skb = esp_skb_alloc(dev, esp_sk->raddr, data_len, allocation, err);
	if (!skb)
		goto out;

	/* set skb properties */
	skb->priority = sk->sk_priority;

	/* Copy esp header into skb */
	esp_hdr = (struct esp_header*) skb_put(skb, ESP_HLEN);
	esp_hdr->dst_port  = esp_sk->rport;
	esp_hdr->src_port  = esp_sk->lport;
	esp_hdr->length    = htons(data_len);
	esp_hdr->pkt_seq   = htons(esp_sk->send_seq);
	esp_hdr->ack_seq   = htons(esp_sk->recv_seq);
	esp_hdr->flags     = flags;

	/* init socket buffer control data */
	ESP_SKB_CB(skb)->pkt_seq = esp_sk->send_seq;
	ESP_SKB_CB(skb)->ack_seq = esp_sk->recv_seq;
	ESP_SKB_CB(skb)->sent_ts = ESP_TIMESTAMP;
	ESP_SKB_CB(skb)->timeo   = ESP_TIMESTAMP + esp_sk->rto;

out:
	return skb;
}

/*
 * Schedules an RRQ to be sent later.
 */
static inline int esp_rrq_schedule(struct sock *sk)
{
	struct esp_opt *esp_sk = ESP_SK(sk);
	int err = 0;

	/* always later, send RRQs as seldom as possible */
	sk_reset_timer(sk, &esp_sk->rrq_timer, ESP_TIMESTAMP + (esp_sk->srtt << 1));

	return err;
}


/*
 * convenience function for enqueuing an skb into the retransmission queue
 */
static inline void esp_skb_enqueue_rtx(struct sock *sk, struct sk_buff *skb)
{
	struct esp_opt *esp_sk = ESP_SK(sk);

	/* update the socket's send sequence number */
	++esp_sk->send_seq;

	/* enqueue the buffer for possible retransmission */
	__skb_queue_tail(&sk->sk_write_queue, skb);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
	sk->sk_wmem_queued += skb->truesize;
	sk_mem_charge(sk, skb->truesize);
#else
	sk_charge_skb(sk, skb);
#endif

	if (!sk->sk_send_head)
		sk->sk_send_head = sk->sk_write_queue.next;

	PR_DEBUG("esp socket %p rtx enqueue "
	         "(tx mem %d/%d, rtx queue len %d mem %d)", sk,
	         atomic_read(&sk->sk_wmem_alloc), sk->sk_sndbuf,
	         sk->sk_write_queue.qlen, sk->sk_wmem_queued);

	/* (re)start retransmission timer */
	if (!timer_pending(&esp_sk->rtx_timer))
		esp_rtx_timer_reset(sk, ESP_LONG_TIMEOUT);
}


#endif
