/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

#include "enet_socket.h"
#include "enet_stream_output.h"
#include "enet_stream_cong.h"
#include "enet_stream_common.h"
#include "enet_compat.h"

/*
 * SENDMSG handler
 * (this means here is where the data comes out of userspace for our joy)
 */
int sock_esp_sendmsg(struct kiocb *iocb, struct socket *sock,
                     struct msghdr *msg, size_t len)
{
	struct sock *sk = sock->sk;
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct net_device *dev;
	struct sk_buff *skb = NULL;
	long timeo;
	int copied;
	char flags = 0;
	int err = 0;
	ENET_PROFILING_DECLARE(unsigned long long time[3]);

	/* take sendmsg() start time */
	ENET_PROFILING_TIMESTAMP(&time[0]);

	lock_sock(sk);

	timeo = sock_sndtimeo(sk, msg->msg_flags & MSG_DONTWAIT);

	/* wait for a connection to finish. */
	if (!SK_STATE(sk, ESPF_ESTABLISHED | ESPF_CLOSE_WAIT))
		if ((err = sk_stream_wait_connect(sk, &timeo)) != 0)
			goto out;

	clear_bit(SOCK_ASYNC_NOSPACE, &sock->flags);

	if (sk->sk_err || (sk->sk_shutdown & SEND_SHUTDOWN)) {
		err = sk_stream_error(sk, msg->msg_flags, -EPIPE);
		goto out;
	}

	PR_DEBUG("socket %p sending %d bytes to "NETH_FORMAT"-%d via if %d",
	         sk, (int)len, NETH_HEX(esp_sk->raddr),
	         ntohs(esp_sk->rport), esp_sk->ifindex);

	/* get device for status and message size check */
	dev = dev_get_by_index(INIT_NET_ARG esp_sk->ifindex);
	if (!dev) {
		PR_ERR("device %d not found", esp_sk->ifindex);
		err = -ENODEV;
		goto out;
	}

	/* check if device is UP */
	if (!(dev->flags & IFF_UP)) {
		PR_ERR("device %s is DOWN", dev->name);
		err = -ENETDOWN;
		SK_ERR_REPORT(sk, err);
		goto out_put;
	}

	if (!esp_sk->trans_announced) {
		/* check if the TXF frame was already sent out */
		if (sk->sk_send_head) {
			/* Good, it wasn't - clear the TXF flag on the last packet */
			PR_DEBUG("Clearing TXF flag on last packet on queue\n");
			WARN_ON(!sk->sk_write_queue.prev);
			WARN_ON(sk->sk_write_queue.prev == (struct sk_buff*)&sk->sk_write_queue);
			WARN_ON(!(ESP_HDR(sk->sk_write_queue.prev)->flags & ESP_FLAG_TXF));
			ESP_HDR(sk->sk_write_queue.prev)->flags &= ~ESP_FLAG_TXF;
		} else {
			/* Tell the reciever we're about to send some data using a new
			 * TXS flagged packet.
			 */
			flags |= ESP_FLAG_TXS;
			/* the next packet is about to carry the TXS flag */
			esp_sk->announce_seq = esp_sk->send_seq;
			esp_rtx_timer_reset(sk, ESP_LONG_TIMEOUT);
		}

		esp_sk->trans_announced = 1;
	}

	/* we may send data */
	copied = 0;

	/*   printk("len: %d\n", len); */

	while (copied < len) {
		/* max. amount of data to copy now */
		int copy;

		skb = sk->sk_write_queue.prev;

		/* packet was already sent out */
		if (!sk->sk_send_head ||
		                /* packet is already mss-sized */
		                ((dev->mtu - ESP_HLEN) - skb->len) < 0) {

alloc_skb:
			copy = min_t(int, len - copied, dev->mtu - ESP_HLEN);

			if (!sk_stream_memory_free(sk))
				goto wait_for_memory;

			if ((copied + copy) == len) {
				/* this will be the last packet for now */
				flags |= ESP_FLAG_TXF;
				esp_sk->trans_announced = 0;
			}

			skb = esp_skb_alloc_sk(sk, dev, copy, flags, sk->sk_allocation, &err);
			if (!skb)
				goto wait_for_memory;

			/*printk ("new skb: %d (room %d)\n", copy, skb_tailroom(skb)); */


			/* flags were copied over */
			flags = 0;

			/* charge the skb to the socket */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
			sk->sk_wmem_queued += skb->truesize;
			sk_mem_charge(sk, skb->truesize);
#else
			sk_charge_skb(sk, skb);
#endif

			/* add this skb to the tail of send queue */
			__skb_queue_tail(&sk->sk_write_queue, skb);

			/* maybe we got a new send_head? */
			if (!sk->sk_send_head)
				sk->sk_send_head = skb;

			/* next packet should get next seq. number */
			++esp_sk->send_seq;

			err = memcpy_fromiovec(skb_put(skb, copy), msg->msg_iov, copy);
			if (err) goto do_fault;

			goto data_copied;
		}

		/* try to append data */

		if (skb_tailroom(skb) > 0) {
			const int off = skb->len;
			copy = min_t(int, dev->mtu - ESP_HLEN - ntohs(ESP_HDR(skb)->length), len - copied);
			copy = min_t(int, skb_tailroom(skb), copy);
			/*         printk("append: %d\n", copy); */
			if (copy == 0) goto alloc_skb;
			err = memcpy_fromiovec(skb_put(skb, copy), msg->msg_iov, copy);

			if (err) {
				__skb_trim(skb, off);
				goto do_fault;
			}

			/* the length changed */
			/* printk("skb len: %d\n", skb->len); */
			ESP_HDR(skb)->length = htons(ntohs(ESP_HDR(skb)->length) + copy);
		} else goto alloc_skb;

data_copied:

		/* maybe we should push some data out */
		esp_push_one_frame(sk, 0);

		copied += copy;

		if (copied == len)
			goto out_sent;

		continue;

wait_for_memory:
		PR_DEBUG("waiting for memory in sendmsg on socket %p", sk);
		set_bit(SOCK_ASYNC_NOSPACE, &sock->flags);
		set_bit(SOCK_NOSPACE, &sock->flags);

		//printk("writer sleeps at %lu\n", ESP_TIMESTAMP);
		err = sk_stream_wait_memory(sk, &timeo);

		if (err) {
			//printk("writer timed out at %lu\n", ESP_TIMESTAMP);
			goto out_sent;
		}
		//printk("writer was woke up at %lu.\n", ESP_TIMESTAMP);
	}

	goto out_sent;

do_fault:
	/* we had a problem appending data */
	PR_DEBUG("there was a problem appending data...");
	if (!skb->len) {
		if (sk->sk_send_head == skb)
			sk->sk_send_head = NULL;
		__skb_unlink(skb, &sk->sk_write_queue);
		sk_wmem_free_skb(sk, skb);
	}

out_sent:
	/* nearly all send*() implementations return the bytes just sent */
	if (copied > 0)
		err = copied;

out_put:
	dev_put(dev);

out:

	if (flags & ESP_FLAG_TXS) {
		/* we failed to enqueue the first packet of a new transmission */
		esp_sk->trans_announced = 0;
	}

	if (flags & ESP_FLAG_TXF) {
		/* same as above for the last packet */
		esp_sk->trans_announced = 1;
	}

	release_sock(sk);

	/* take sendmsg() end time */
	ENET_PROFILING_TIMESTAMP(&time[2]);
	PR_PROFILING("SENDMSG %p %d bytes %llu - %llu - %llu (%llu, %llu)",
	             sk, len, time[0], time[1], time[2],
	             time[1] - time[0], time[2] - time[0]);

	PR_DEBUG("sock %p sent out %d bytes of %lu", sk, err, len);
	return err;
}

/**
 * Pushes out the next pending frame. It obeys sending limitations.
 * @param fix_empty_sbuff if it is allowed to add the TXF flag
 *    if the send buffer becomes empty.
 * @return 1 if a frame was sent, 0 if CA banned it, < 0 for errors
 */
int esp_push_one_frame(struct sock *sk, int fix_empty_sbuff)
{
	struct sk_buff *skb = sk->sk_send_head;
	int err = 0;

	if (skb) {
		/* ask CA if it's ok to send the frame */
		if (!esp_cwnd_test(sk, skb))
			return 0;

		if (fix_empty_sbuff &&
		                ESP_SK(sk)->trans_announced &&
		                (skb->next == (struct sk_buff *)&sk->sk_write_queue)) {
			/* that's a bit strange. our send buffer is about to
			 * become empty, but a trans is still announced. we
			 * fix this here.
			 */

			ESP_HDR(skb)->flags |= ESP_FLAG_TXF;
			ESP_SK(sk)->trans_announced = 0;
		}

		/* send out a _clone_ of the skb, as it remains enqueued */
		if ((err = esp_skb_xmit(sk, skb_clone(skb, GFP_ATOMIC))) < 0) {
			ESP_SKB_CB(skb)->xmits++;
			return err;
		}

		/* advance send head */
		sk->sk_send_head = skb->next;
		if (sk->sk_send_head == (struct sk_buff *)&sk->sk_write_queue)
			sk->sk_send_head = NULL;

	} else {
		return -ENOBUFS;
	}

	return 1;
}

/**
 * Pushes out any pending frames held back due to congestion.
 * This function obeys congestion limits.
 * @return the number of pushed frames
 */
int esp_push_pending_frames(struct sock *sk)
{
	int cnt = 0;
	while (esp_push_one_frame(sk, 1) > 0) cnt++;
	PR_DEBUG("Pushed %d pending frames on socket %p (%d on queue).",
	         cnt, sk, sk->sk_write_queue.qlen);
	return cnt;
}

/*
 * Convenience function for replying RST to a received skb
 */
int esp_send_rst(struct sk_buff *src_skb)
{
	struct ethhdr *src_mac = (struct ethhdr *) skb_mac_header(src_skb);
	struct esp_header *src_esp_hdr = ESP_HDR(src_skb);
	struct sk_buff *dst_skb;
	struct esp_header *dst_esp_hdr;
	int err = 0;

	/* never reply RST to a RST... */
	if (src_esp_hdr->flags & ESP_FLAG_RST)
		goto out;

	/* allocate the socket buffer */
	dst_skb = esp_skb_alloc(SKB_INPUT_DEV(src_skb), &src_mac->h_source, 0, GFP_ATOMIC, &err);

	if (!dst_skb)
		goto out;

	/* fill in the esp header fields */
	dst_esp_hdr = (struct esp_header*) skb_put(dst_skb, ESP_HLEN);
	dst_esp_hdr->dst_port = src_esp_hdr->src_port;
	dst_esp_hdr->src_port = src_esp_hdr->dst_port;
	dst_esp_hdr->length   = htons(0);
	dst_esp_hdr->pkt_seq  = src_esp_hdr->ack_seq;
	dst_esp_hdr->ack_seq  = htons(ESP_SKB_CB(src_skb)->pkt_seq + 1);
	dst_esp_hdr->flags    = ESP_FLAG_RST;

	if (!(src_esp_hdr->flags & ESP_FLAG_ACK)) {
		dst_esp_hdr->flags |= ESP_FLAG_ACK;
	}

	PR_DEBUG("esp replying "ESP_FLAGS_FORMAT" to "NETH_FORMAT"-%d",
	         ESP_FLAGS_DEBUG(dst_esp_hdr->flags),
	         NETH_HEX(src_mac->h_source), ntohs(src_esp_hdr->src_port));

	/* send the buffer */
	err = esp_skb_xmit(NULL, dst_skb);

out:
	return err;
}

/*
 * Convenience function for sending a SYN packet
 */
int esp_send_syn(struct sock *sk, struct net_device *dev, gfp_t allocation)
{
	struct sk_buff *skb;
	int flags = ESP_FLAG_SYN;
	int err = 0;

	PR_DEBUG("-> esp_send_syn(sk: %p, net_device: %p, alloc: %d)", sk, dev, allocation);

	/* check whether ACKs are pending (for SYN-ACK) */
	if (ESP_ACK_PENDING(sk)) {
		flags |= ESP_FLAG_ACK;
	}

	/* allocate and prepare a socket buffer containing the SYN packet */
	PR_DEBUG("allocating socket buffer");
	skb = esp_skb_alloc_sk(sk, dev, 0, flags, allocation, &err);
	if (!skb) {
		PR_ERR("esp socket %p failed to allocate "ESP_FLAGS_FORMAT
		       " packet with err %d",
		       sk, ESP_FLAGS_DEBUG(flags), err);
		goto out;
	}

	/* enqueue the skb for retransmission, start rtx timer */
	PR_DEBUG("enqueuing for retransmission");
	esp_skb_enqueue_rtx(sk, skb);

	/* send the buffer */
	//   PR_DEBUG("ximitting for the first time");
	//   err = esp_skb_xmit(sk, skb_clone(skb, allocation));
	//   if (err < 0) {
	//      PR_DEBUG("esp socket %p failed to send "ESP_FLAGS_FORMAT
	//        " packet with err %d",
	//        sk, ESP_FLAGS_DEBUG(flags), err);
	//      goto out;
	//   }

	//   ++ESP_SKB_CB(skb)->xmits;

	esp_push_pending_frames(sk);

	PR_DEBUG("esp socket %p port %d sent "ESP_FLAGS_FORMAT" to "NETH_FORMAT"-%d",
	         sk, ESP_SK(sk)->port, ESP_FLAGS_DEBUG(flags),
	         NETH_HEX(ESP_SK(sk)->raddr), ntohs(ESP_SK(sk)->rport));

out:
	PR_DEBUG("<- esp_send_syn");
	return err;
}

/*
 * Convenience function for sending a FIN packet
 */
int esp_send_fin(struct sock *sk, gfp_t allocation)
{
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct net_device *dev;
	struct sk_buff *skb;
	unsigned char flags = ESP_FLAG_FIN;
	int err = 0;

	dev = dev_get_by_index(INIT_NET_ARG esp_sk->ifindex);
	if (!dev) {
		err = -ENODEV;
		goto out;
	}

	/* check whether ACKs are pending (for backpacking) */
	if (ESP_ACK_PENDING(sk)) {
		flags |= ESP_FLAG_ACK;
	}

	/* allocate the FIN socket buffer */
	skb = esp_skb_alloc_sk(sk, dev, 0, flags, allocation, &err);
	if (!skb) {
		PR_ERR("esp socket %p failed to allocate "ESP_FLAGS_FORMAT
		       " packet with err %d",
		       sk, ESP_FLAGS_DEBUG(flags), err);
		goto out_put;
	}

	/* enqueue the skb for retransmission, re-start rtx timer */
	esp_skb_enqueue_rtx(sk, skb);

	/* try to get it out */
	esp_push_pending_frames(sk);

	/* send the buffer */
	//   err = esp_skb_xmit(sk, skb_clone(skb, allocation));
	//   if (err < 0) {
	//      PR_DEBUG("esp socket %p failed to send "ESP_FLAGS_FORMAT
	//        " packet with err %d",
	//        sk, ESP_FLAGS_DEBUG(flags), err);
	//      goto out_put;
	//   }

	//   ++ESP_SKB_CB(skb)->xmits;

out_put:
	dev_put(dev);
out:
	return err;
}

/**
 * Enqueues ACK at tail, sends the ACK at head. Should be
 * called only if the socket can bear the data to expect
 * (no memory pressure).
 *
 * @param flags extra flags that should be set on the ACK to be
 *              enqueued (most likely ESP_FLAG_RRQ; but anything
 *              goes) No need to put ESP_FLAG_ACK here, as that's
 *              always set.
 */
int esp_enqueue_ack(struct sock *sk, int flags)
{
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct sk_buff *skb = NULL;
	struct net_device *dev_xmit, *dev_enq;
	int err = 0;

	flags |= ESP_FLAG_ACK;

	PR_DEBUG("Socket %p enqueues ACK (recv_count:%d)",
	         sk, atomic_read(&esp_recv_count));

#if ENET_DEBUG
	/* there should be no ACK on queue for a socket which calls this
	 * function
	 */
	spin_lock_bh(&esp_ack_queue.lock);

	__esp_walk_ack_queue(skb) {
		if (skb->sk == sk) {
			PR_ERR("A socket called esp_enqueue_ack while having an ACK on queue!");
			spin_unlock_bh(&esp_ack_queue.lock);
			return 1;
		}
	}

	spin_unlock_bh(&esp_ack_queue.lock);

	/* and it shouldn't be under memory pressure either */
	WARN_ON(esp_sk->rcv_mem_pressure);

#endif

	if ((dev_enq = dev_get_by_index(INIT_NET_ARG esp_sk->ifindex)) == NULL) {
		PR_ERR("Hey, the device isn't there!");
		err = -ENODEV;
		goto out;
	}

	/* generate the ACK this socket wants to send */
	skb = esp_skb_alloc_sk(sk, dev_enq, 0, flags, GFP_ATOMIC, &err);
	if (!skb)
		goto out_put;

	/* charge it to the socket */
	skb_set_owner_w(skb, sk);

	/* This ACK will be sent sooner or later, so we record
	 * this fact with the socket.
	 */
	esp_sk->ack_seq = esp_sk->recv_seq;

	/* if there is only one receiver we can simply xmit the ACK */
	if (atomic_read(&esp_recv_count) < 2) {
		PR_DEBUG("Skipping ACK enqueuing for socket %p.\n", sk);
		goto out_xmit;
	}

	/* we can stop the RRQ timer */
	esp_rrq_clear(sk);

	spin_lock_bh(&esp_ack_queue.lock);

	PR_DEBUG("Socket %p enqueues ACK (recv_count:%d, ack queue len: %d)\n",
	         sk, atomic_read(&esp_recv_count), esp_ack_queue.qlen);

	__skb_queue_tail(&esp_ack_queue, skb);

	/* now we can easily decide if we really want to xmit an
	 * ACK or if the queue had to grow.
	 */
	if (esp_ack_queue.qlen < atomic_read(&esp_recv_count)) {
		/* the queue should have grown. it has. we're done. */

		PR_DEBUG("Socket %p left ACK queue grow (recv_count: %d, qlen: %d).\n",
		         sk, atomic_read(&esp_recv_count), esp_ack_queue.qlen);

		spin_unlock_bh(&esp_ack_queue.lock);
		goto out_put;
	}

	skb = __skb_dequeue(&esp_ack_queue);

	spin_unlock_bh(&esp_ack_queue.lock);

out_xmit:
	/* now skb points to the ACK we'd like to xmit */
	if ((dev_xmit = dev_get_by_index(INIT_NET_ARG ESP_SK(skb->sk)->ifindex)) == NULL) {
		err = -ENODEV;
		PR_ERR("Uh, the device went offline while I had an ACK to xmit on queue.");
		kfree_skb(skb);
		goto out_put;
	}

	PR_DEBUG("We xmit an ACK for socket %p.\n", skb->sk);

	/* the ACK will be sent, start RRQ timer */
	ESP_RESET_RRQ_TIMER(skb->sk, esp_round_trip_time);

	/* it's likely we received some more packets since the ACK
	 * was enqueued. so we adjust the ack_seq for that.
	 */

	ESP_HDR(skb)->ack_seq = htons(ESP_SK(skb->sk)->recv_seq);

	err = dev_queue_xmit(skb);
	if (err != NET_XMIT_SUCCESS) {
		PR_DEBUG("esp socket %p failed to enqueue skb containing an ACK for "
		         "sending with errno %d", sk, err);
		if (err > 0)
			err = net_xmit_errno(err);
		if (err != 0)
			goto out_put;
	}

	dev_put(dev_xmit);

out_put:
	dev_put(dev_enq);

out:
	return err;
}

int esp_send_rrq(struct sock *sk)
{
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct net_device *dev;
	struct sk_buff *skb;
	unsigned char flags = ESP_FLAG_ACK | ESP_FLAG_RRQ;
	int err = 0;

	dev = dev_get_by_index(INIT_NET_ARG esp_sk->ifindex);
	if (!dev) {
		err = -ENODEV;
		goto out;
	}

	/* allocate the RRQ socket buffer */
	skb = esp_skb_alloc_sk(sk, dev, 0, flags, GFP_ATOMIC, &err);
	if (!skb)
		goto out_put;

	/* send the RRQ */
	err = esp_skb_xmit(sk, skb);
	if (err) {
		PR_DEBUG("RRQ transmission failed on esp socket %p err %d",
		         sk, err);
		goto out_put;
	}

	/* we always send an ACK as well */
	esp_sk->ack_seq = esp_sk->recv_seq;
	esp_sk->stats.arq_tx++;

	PR_DEBUG("Socket %p sent RRQ %d.", sk, esp_sk->recv_seq);

out_put:
	dev_put(dev);
out:
	return err;
}

/*
 * Convenience function for sending an immediate ACK packet
 */
int esp_send_ack(struct sock *sk)
{
	struct esp_opt *esp_sk = ESP_SK(sk);
	struct net_device *dev;
	struct sk_buff *skb;
	unsigned char flags = ESP_FLAG_ACK;
	int err = 0;

	dev = dev_get_by_index(INIT_NET_ARG esp_sk->ifindex);
	if (!dev) {
		err = -ENODEV;
		goto out;
	}

	/* allocate the ACK socket buffer */
	skb = esp_skb_alloc_sk(sk, dev, 0, flags, GFP_ATOMIC, &err);
	if (!skb)
		goto out_put;

	/* send the ACK */
	err = esp_skb_xmit(sk, skb);
	if (err) {
		PR_DEBUG("ack transmission failed on esp socket %p err %d",
		         sk, err);
		goto out_put;
	}

	esp_sk->ack_seq = esp_sk->recv_seq;

	PR_DEBUG("esp socket %p sent ack %d", sk, esp_sk->recv_seq);

out_put:
	dev_put(dev);
out:
	return err;
}

/*
 * Convenience function for SKB transmission
 * (provide a clone of the original sockt buffer if it is enqueued
 *  e.g. in the retransmission queue)
 */
int esp_skb_xmit(struct sock *sk, struct sk_buff *skb)
{
	int err = 0;

	if (skb) {
		/* charge the skb to the socket */
		if (sk)
			skb_set_owner_w(skb, sk);

		/* queue the skb for transmission */
		err = dev_queue_xmit(skb);
		if (err != NET_XMIT_SUCCESS) {
			PR_DEBUG("esp socket %p failed to enqueue skb for "
			         "sending with errno %d", sk, err);
			if (err > 0)
				err = net_xmit_errno(err);
			if (err != 0)
				goto out;
		}

		PR_DEBUG("esp socket %p sent frame (ts %lu)",
		         sk, ESP_TIMESTAMP);

	} else {
		err = -ENOBUFS;
	}

out:
	return err;
}
