/*
 * Copyright (c) 2006 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 * Copyright (c) 2006 The Technical University of Chemnitz. All 
 *                    rights reserved.
 *
 * Author(s): (alphabetically)
 *            Sven Eckelmann
 *            Torsten Hoefler 
 *            Mirko Reinhard 
 *            Matthias Treydte 
 */

#include <linux/version.h>
#include <linux/proc_fs.h>
#include <linux/module.h>
#include <linux/seq_file.h>
#include <net/sock.h>

#include "enet_stream_stats.h"
#include "enet_module.h"
#include "enet_stream.h"
#include "enet_stream_cong.h"
#include "enet_stream_common.h"

/**
 * Moves the given iterator to the given position.
 */
static __inline__ void esp_seq_idx(struct htable_iterator *it, loff_t off)
{
	struct sock *sk;
	struct hlist_node *node;
	int i;

	for (i = 0; i < ESP_BHTABLE_SIZE; i++)
		sk_for_each_bound(sk, node, esp_bhash + i) {
		if (!off--) {
			/* element found */
			it->bucket = i;
			it->sk = sk;
			return;
		}
	}

	/* element not found, make iterator invalid */
	it->bucket = 0;
	it->sk = NULL;
}

/**
 * Moves the given iterator to the next valid position
 * within the hashtable. If there are no elements left
 * an invalid iterator is returned. If an invalid iterator
 * is given, it's moved to the first element in the table.
 */
static __inline__ void move_next(struct htable_iterator *it)
{
	if (! ITER_VALID(it)) {
		/* find the first used entry */

		for (it->bucket = 0; it->bucket < ESP_EHTABLE_SIZE; it->bucket++) {
			it->sk = sk_head(esp_bhash + it->bucket);
			if (it->sk) return;
		}

		/* If we fall through here it->sk == NULL,
		 * which means the iterator is invalid. That's
		 * what we want.
		 */
	} else {
		/* move to the next used entry */
		it->sk = sk_next(it->sk);
		if (it->sk) return;

		/* move to the next bucket */
		do {
			it->sk = sk_head(esp_bhash + ++(it->bucket));
			if (it->sk) return;
		} while (it->bucket < ESP_BHTABLE_SIZE);
	}
}

static void *esp_seq_start(struct seq_file *seq, loff_t *pos)
__acquires(esp_bhash_lock)
{
	struct htable_iterator *it = kmalloc(sizeof(struct htable_iterator), GFP_KERNEL);
	ITER_INIT(it);
	read_lock(&esp_bhash_lock);

	if (*pos) {
		esp_seq_idx(it, *pos - 1);

		if (! ITER_VALID(it)) {
			kfree(it);
			return NULL;
		}
	}

	return it;
}

static void *esp_seq_next(struct seq_file *seq, void *v, loff_t *pos)
{
	++*pos;
	move_next((struct htable_iterator *)v);

	if (! ITER_VALID(((struct htable_iterator *)v))) {
		/* reached end of elements */
		kfree(v);
		return NULL;
	}

	return v;
}

static void esp_seq_stop(struct seq_file *seq, void *v)
__releases(esp_bhash_lock)
{
	read_unlock(&esp_bhash_lock);
	/* in case we've been stopped before using all elements */
	if (v) kfree(v);
}

static int esp_seq_show(struct seq_file *s, void *v)
{
	struct htable_iterator *it = (struct htable_iterator *)v;

	if (ITER_BEFORE_FIRST(it)) {
		seq_printf(s, "Global stats:\n");
		seq_printf(s, "recv_count:    %d\n", atomic_read(&esp_recv_count));
		seq_printf(s, "sockets alive: %d\n", atomic_read(&esp_skcount));
		seq_printf(s, "\n");
		seq_printf(s, "sk       DstAddr           LPort RefCnt IfIdx TXCnt RTXCnt state\n");
	} else {
		const struct esp_opt *opt = ESP_SK(it->sk);
		const struct esp_stats *stats = &opt->stats;

		sock_hold(it->sk);

		seq_printf(s, "%p ", it->sk);
		seq_printf(s, NETH_FORMAT, NETH_HEX(opt->raddr));
		seq_printf(s, " %-5d ", ntohs(opt->lport));
		seq_printf(s, "%-6d ", atomic_read(&it->sk->sk_refcnt));
		seq_printf(s, "%-5d ", opt->ifindex);
		seq_printf(s, "%-5d ", stats->pkt_tx);
		seq_printf(s, "%-6d ", stats->pkt_rtx);
		seq_printf(s, esp_statenames[it->sk->sk_state]);

		seq_printf(s, "\n");

		sock_put(it->sk);
	}

	return 0;
}

static struct seq_operations esp_seq_ops = {
	.start = esp_seq_start,
	.next  = esp_seq_next,
	.stop  = esp_seq_stop,
	.show  = esp_seq_show,
};


static int esp_seq_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &esp_seq_ops);
}


/**
 * file operations registration structure
 */
struct file_operations esp_proc_fops = {
	.owner   = THIS_MODULE,
	.open    = esp_seq_open,
	.read    = seq_read,
	.llseek  = seq_lseek,
	.release = seq_release,
};

